# %% Imports
import numpy as np
import pandas as pd
from test_functions import *

# %% Helper Functions

def vectors_to_formatted_strings(df):
    df = df.copy()
    def format_vector(v):
        return np.array2string(v, floatmode = 'unique', 
            max_line_width = sys.maxsize, threshold = sys.maxsize, separator = ',')
    for name, values in df.select_dtypes('object').iteritems(): 
        s = pd.Series(map(format_vector, values))
        df[name] = s.str.replace('[\[\]" "]', '', regex = True)
    return df

def vectors_to_unrolled_scalars(df, remove_vectors=True):
    df = df.copy()
    for col_name, values in df.select_dtypes('object').iteritems():
        data = np.vstack(values).transpose()
        col_names = [col_name + "_" + str(i) for i in range(data.shape[1])]
        for col, new_name in zip(data, col_names):
            idx = df.columns.get_loc(col_name)
            df.insert(idx, new_name, col)
        if remove_vectors: 
            df.drop(columns = col_name, inplace = True)
    return df

def vectors_to_aggregated_scalars(df, reduction_ops=[np.sum, np.mean, np.std, np.var, np.median, np.min, np.max], remove_vectors=True):
    df = df.copy()
    def cov(x, y):
        x, y = np.vstack(x), np.vstack(y)
        x_m, y_m = np.mean(x, axis=1, keepdims=True), np.mean(y, axis=1, keepdims=True)
        x_d, y_d = x - x_m, y - y_m
        s = np.sum(x_d * y_d, axis=1)
        return s / x.shape[1]
    if 'v1' in df.columns and 'v2' in df.columns:
        v1_v2_cov = cov(df['v1'], df['v2'])
        df.insert(0, 'v1_v2_cov', v1_v2_cov)
    for col_name, values in df.select_dtypes('object').iteritems():
        data = np.vstack(values)
        for reduce_op in reduction_ops:
            reduced_vector = reduce_op(data, axis=1)
            idx = df.columns.get_loc(col_name)
            new_name = f'{col_name}_{reduce_op.__name__}'
            df.insert(idx, new_name, reduced_vector)
        if remove_vectors: 
            df.drop(columns=col_name, inplace=True)
    return df

# %% Generator Functions

def generate_data_frame(f, rows, v_len, seed):
    rng = np.random.default_rng(seed)
    data = f(rng, rows, v_len)
    # wrap vector data in lists
    data = { name : col if col.ndim == 1 else list(col) for name, col in data.items() }
    return pd.DataFrame(data)

def build_test(f, num_rows=1000, v_len=20, seed=None):
    data = generate_data_frame(f, num_rows, v_len, seed)
    name = f.__name__

    formatted_vector_data = vectors_to_formatted_strings(data)
    formatted_vector_data.to_csv(f"data/{name}.csv", index=False, sep=";")

    unrolled_data = vectors_to_unrolled_scalars(data)
    unrolled_data.to_csv(f"data/unrolled_{name}.csv", index=False, sep=";")

    pre_aggregated_data = vectors_to_aggregated_scalars(data)
    pre_aggregated_data.to_csv(f"data/pre_aggregated_{name}.csv", index=False, sep=";")


# %% Main
if __name__ == "__main__":
    build_test(test_A_01, 1000, 20, 101)
    build_test(test_A_02, 1000, 20, 102)
    build_test(test_A_03, 1000, 20, 103)
    build_test(test_A_04, 1000, 20, 104)

    build_test(test_B_01, 1000, 20, 201)
    build_test(test_B_02, 1000, 20, 202)
    build_test(test_B_03, 1000, 20, 203)
    build_test(test_B_04, 1000, 20, 204)
    build_test(test_B_05, 1000, 20, 205)
    build_test(test_B_06, 1000, 20, 206)
