import numpy as np

# uniform scalar
def Us(rng, rows, lower, upper):
    lower_mean, lower_std = lower if isinstance(lower, tuple) else (lower, 0)
    upper_mean, upper_std = upper if isinstance(upper, tuple) else (upper, 0)
    l = rng.normal(lower_mean, lower_std, size=rows)
    u = rng.normal(upper_mean, upper_std, size=rows)
    return rng.uniform(l, u, size=rows)

# uniform vector
def Uv(rng, rows, v_len, lower, upper):
    lower_mean, lower_std = lower if isinstance(lower, tuple) else (lower, 0)
    upper_mean, upper_std = lower if isinstance(upper, tuple) else (upper, 0)
    l = rng.normal(lower_mean, lower_std, size=(rows, 1))
    u = rng.normal(upper_mean, upper_std, size=(rows, 1))
    return rng.uniform(l, u, size=(rows, v_len))

def Uv_mr(rng, rows, v_len, mean, range):
    if isinstance(mean, tuple): mean = rng.uniform(mean[0], mean[1], size=rows)
    if isinstance(range, tuple): range = rng.uniform(range[0], range[1], size=rows)
    lower, upper = mean - range/2, mean + range/2
    return rng.uniform(lower.reshape(-1, 1), upper.reshape(-1, 1), size=(rows, v_len))

def Uv_ms(rng, rows, v_len, mean, sigma):
    if isinstance(sigma, tuple): return Uv_mr(rng, rows, v_len, mean, tuple([s * np.sqrt(12) for s in sigma]))
    return Uv_mr(rng, rows, v_len, mean, sigma * np.sqrt(12))


def test_A_01(rng, rows, v_len):
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    y = 2.5 * np.mean(v1, axis=1)
    return { 'v1': v1, 'y': y }

def test_A_02(rng, rows, v_len):
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    y = np.std(v1, axis=1) - 2
    return { 'v1': v1, 'y': y }

def test_A_03(rng, rows, v_len):
    x1 = Us(rng, rows, lower=+1, upper=+4)
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    y = 2.5 * x1 + np.mean(v1, axis=1) + 2
    return { 'x1': x1, 'v1': v1, 'y': y }

def test_A_04(rng, rows, v_len):
    x1 = Us(rng, rows, lower=+1, upper=+4)
    x2 = Us(rng, rows, lower=-8, upper=-4)
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    v2 = Uv_ms(rng, rows, v_len, mean=(10, 20), sigma=(4, 8))
    y = (x1 * np.var(v1, axis=1))/3 - 3*np.mean(v2, axis=1) / x2
    return { 'x1': x1, 'x2': x2, 'v1': v1, 'v2': v2, 'y': y }


def test_B_01(rng, rows, v_len):
    x1 = Us(rng, rows, lower=+1, upper=+4)
    x2 = Us(rng, rows, lower=-8, upper=-4)
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    v2 = Uv_ms(rng, rows, v_len, mean=(10, 20), sigma=(4, 6))
    y = x1 * np.mean(v1 + v2, axis=1)
    return { 'x1': x1, 'x2': x2, 'v1': v1, 'v2': v2, 'y': y }

def test_B_02(rng, rows, v_len):
    x1 = Us(rng, rows, lower=+1, upper=+4)
    x2 = Us(rng, rows, lower=-8, upper=-4)
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    v2 = Uv_ms(rng, rows, v_len, mean=(10, 20), sigma=(4, 6))
    y = x1 * np.mean(v1 * v2, axis=1)
    return { 'x1': x1, 'x2': x2, 'v1': v1, 'v2': v2, 'y': y }

def test_B_03(rng, rows, v_len):
    x1 = Us(rng, rows, lower=+1, upper=+4)
    x2 = Us(rng, rows, lower=-8, upper=-4)
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    v2 = Uv_ms(rng, rows, v_len, mean=(10, 20), sigma=(4, 6))
    y = x1 * np.std(v1 + v2, axis=1)
    return { 'x1': x1, 'x2': x2, 'v1': v1, 'v2': v2, 'y': y }

def test_B_04(rng, rows, v_len):
    x1 = Us(rng, rows, lower=+1, upper=+4)
    x2 = Us(rng, rows, lower=-8, upper=-4)
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    v2 = Uv_ms(rng, rows, v_len, mean=(10, 20), sigma=(4, 6))
    y = x1 * np.std(v1 * v2, axis=1)
    return { 'x1': x1, 'x2': x2, 'v1': v1, 'v2': v2, 'y': y }

def test_B_05(rng, rows, v_len):
    x1 = Us(rng, rows, lower=+1, upper=+4)
    x2 = Us(rng, rows, lower=-8, upper=-4)
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    v2 = Uv_ms(rng, rows, v_len, mean=(10, 20), sigma=(4, 6))
    v3 = Uv_ms(rng, rows, v_len, mean=(2, 4), sigma=(0.05, 0.15))
    y = x1 * np.mean((v1 + 2*v3)/(0.5*v2), axis=1)
    return { 'x1': x1, 'x2': x2, 'v1': v1, 'v2': v2, 'v3': v3, 'y': y }

def test_B_06(rng, rows, v_len):
    x1 = Us(rng, rows, lower=+1, upper=+4)
    x2 = Us(rng, rows, lower=-8, upper=-4)
    v1 = Uv_ms(rng, rows, v_len, mean=(4, 8), sigma=(2, 4))
    v2 = Uv_ms(rng, rows, v_len, mean=(10, 20), sigma=(4, 6))
    v3 = Uv_ms(rng, rows, v_len, mean=(2, 4), sigma=(0.05, 0.15))
    y = x1 * np.std((v1 + 2*v3)/(0.5*v2), axis=1)
    return { 'x1': x1, 'x2': x2, 'v1': v1, 'v2': v2, 'v3': v3, 'y': y }
